#ifndef __XBEEPACKETS_H
#define __XBEEPACKETS_H

//define API_MODE_2 if you want to use escape characters.  Otherwise it will assume XBee API mode 1.
#define API_MODE_2

#include "main.h"
#include <uart.h>

//peng, 8/6
#include "e_motors_swarm.h"
//peng

//This union is used to split a 32-bit Float into 4 bytes	
union packetFloat{
	float dataFloat;
	unsigned char dataChar[DATATYPE_BYTELENGTH];
};

//The most recent received packet is stored into this struct
volatile struct IN_PACKET
{
	unsigned int length;
	unsigned char API_ID;
	unsigned int source;
	unsigned char RSSI;
	unsigned char options;
	union packetFloat data[DATA_ARRAY_LENGTH+ADDITIONAL_NUMS];                //peng, 8/6
	unsigned char checksum;
}InPacket;

#define TIMEOUT_COUNT 5000
#define WAIT_TIMEOUT 	timeout=0;\
	while (timeout<TIMEOUT_COUNT && !DataRdyUART2() ){\
		timeout++;\
	}\
	if (timeout>=TIMEOUT_COUNT){\
		return -1;\
	}


//Private Variables:
static unsigned char checksum=0;
static unsigned int timeout=0;
static unsigned char QUERIED_ID=0xFF;


//Public Functions:
//unsigned char getMY_ID( void );
void assemblePacket( float rawData[], unsigned int destination );  //assembles and sends packets 
int readPacket( void );
void flushBufferUART2(void);

//Private Functions:

//read a byte from buffer, apply escapes if needed.
static unsigned char readByte(void);	

//set up packet headers in outPacket raw packet array
static void initPacket(unsigned char outPacket[], unsigned int *indexPtr);	

//load a new byte into outPacket at the index
static void loadByte(unsigned char outPacket[], unsigned char byte, unsigned int *indexPtr);

//sends assembled packet to UART2
static void sendPacket(unsigned char outPacket[], unsigned int Index);



//Public Functions///////////////////////

//retreives XBee radio ID
/*
unsigned char query_MY_ID(void){
	int statusFlag=-1;
	unsigned char tempRCIE;
	unsigned const char outPacket[8]={0x7E, 0x00, 0x04, 0x08, 0x52, 0x4D, 0x59, 0xFF};
	unsigned const char resetPacket[8]={0x7E, 0x00, 0x04, 0x08, 0x52, 0x46, 0x52, 0x0D};

	tempRCIE=IEC1bits.U2RXIE;
	IEC1bits.U2RXIE=0;
	while(QUERIED_ID==255){
		delay(FOSC/5000);
		sendPacket(resetPacket,7);

		while (statusFlag != 1){
			timeout=0;
			flushBufferUART2();
			sendPacket(outPacket, 7);
			while (!DataRdyUART2() && timeout<50000){
				timeout+=1;	
			}
			statusFlag=readPacket();
			delay(FOSC/500);
		}
	}
	IEC1bits.U2RXIE=tempRCIE;
	return QUERIED_ID;
}
*/

//Assembles and sends packet
	
void assemblePacket(float rawData[], unsigned int destination){
	unsigned int i; //i is the index
	unsigned int j;//j is for splitting up signed int into BYTEs
	union packetFloat tempFloatUnion;
	static unsigned char outPacket[90];
	unsigned int index=0;
	unsigned int *indexPtr;
	indexPtr=&index;

//assemble outPacket
	initPacket(outPacket,indexPtr); 
	loadByte(outPacket,0x01,indexPtr);
	loadByte(outPacket,0x01,indexPtr);
	loadByte(outPacket,destination/256,indexPtr);
	loadByte(outPacket,destination%256,indexPtr);
	loadByte(outPacket,0x00,indexPtr);
	
	for (i=0; i<(DATA_ARRAY_LENGTH+ADDITIONAL_NUMS); i++){
		tempFloatUnion.dataFloat=rawData[i];
		//you can only send single bytes at a timeto the serial port, so use a UNION to chop up the 32-bit float.
		for (j=0; j<DATATYPE_BYTELENGTH; j++){	
			loadByte(outPacket,tempFloatUnion.dataChar[DATATYPE_BYTELENGTH-1-j],indexPtr); //read in byte
		}
	}
	loadByte(outPacket,0xFF-checksum,indexPtr);
	sendPacket(outPacket,index);
}

//reads packets
int readPacket( void){
	//////////////SUCCESS RETURN CODES///////////////
	//0: Receive data packet
	//1: Get ID
	//2: TX Status Packet
	//3: Modem Status
	//////////////ERROR CODES/////////////////
	// 
	// -1: Timeout waiting for UART2
	// -2: Start delimiter wrong
	// -3: Checksum Error
	// -4: UART2 not ready at beginning
	// -5: API_ID unidentified
	// -6: Send Packet CCA failuue
	// -7: Modem Status packet failure
	//////////////////////////////////////////
	
	unsigned int i;
	unsigned int j;
	unsigned char tempChar;
	unsigned char verifySum;
	verifySum = 0;

	if(!DataRdyUART2()){
		return -4;
	}
	else
		tempChar=ReadUART2();
		
	if (tempChar != 0x7E){
		return -2;
	}
	
	WAIT_TIMEOUT
	InPacket.length=readByte(); 
	InPacket.length=InPacket.length<<8;
	WAIT_TIMEOUT
	InPacket.length+=readByte();
	WAIT_TIMEOUT
	InPacket.API_ID=readByte();
	switch (InPacket.API_ID){
		case 0x81:
			WAIT_TIMEOUT
			InPacket.source=readByte();
			InPacket.source=InPacket.source<<8;
			WAIT_TIMEOUT
			InPacket.source=readByte()+InPacket.source;
			WAIT_TIMEOUT
			InPacket.RSSI=readByte();
			WAIT_TIMEOUT
			InPacket.options=readByte();
			for (i=0; i<DATA_ARRAY_LENGTH+ADDITIONAL_NUMS; i++){	      
				for (j=0; j<DATATYPE_BYTELENGTH; j++){
					WAIT_TIMEOUT
					InPacket.data[i].dataChar[DATATYPE_BYTELENGTH-1-j]=readByte();	//send out big end first, so that the float will appear
					verifySum+=InPacket.data[i].dataChar[DATATYPE_BYTELENGTH-1-j];	//in order if you examine the raw data.
				}
			}
			WAIT_TIMEOUT
			InPacket.checksum=readByte();
		
			verifySum+=InPacket.API_ID 
					+ InPacket.source 
					+ InPacket.RSSI 
					+ InPacket.options 
					+ InPacket.checksum;
			if (verifySum == 0xFF){
				return 0;
			}
			else{
				return -3;
			}	
			break;
/*
		//This is buggy, but don't need it anyway.
		case 0x88: //Command response, for Query ID	
			verifySum=0x88;
			for (i=1; i<=7; i++){
				WAIT_TIMEOUT
				tempChar=ReadUART2();
				if (tempChar == 0x7D){
					WAIT_TIMEOUT
					tempChar=ReadUART2();
					tempChar=tempChar^0x20;
				}
				verifySum+=tempChar;
				if (i==4 && tempChar==1){
					return -5;
				}
				if (i==6){
					QUERIED_ID=tempChar;
				}
			}
			if (verifySum == 0xFF){
				return 1;
			}
			else{
				return -3;
			}	
			break;
*/
		
		case 0x89:  //send response
			verifySum=0x89;
			for (i=0; i<3; i++){
				WAIT_TIMEOUT
				tempChar=readByte();
				verifySum+=tempChar;
			}
			if (tempChar==2){
				return -7;
			}
			if (verifySum == 0xFF){
				return 2;
			}
			else{
				return -6;
			}	
			break;		
			
		case 0x8A:  //modem status
			verifySum=0x8A;
			for (i=0; i<2; i++){
				WAIT_TIMEOUT
				tempChar=readByte();
				verifySum+=tempChar;
			}	
	
			if (verifySum == 0xFF){
				return 3;
			}
			else{
				return -7;
			}	
			break;		
					
		default:
			return -5;
	}
		
}

//clears incoming buffer until timeout.


void flushBufferUART2(){
	unsigned int timeout=0;
	unsigned char timeoutFlag=0;

	U2STAbits.OERR=0;
	while (timeoutFlag == 0){
		while (timeout<TIMEOUT_COUNT && !DataRdyUART2() ){
			timeout++;
		}
		if (timeout>=TIMEOUT_COUNT){
			timeoutFlag=1;
		}
		else{
			ReadUART2();
			timeout=0;
		}
	}
}



//Private Functions/////////////////////////


static unsigned char readByte(void){
	unsigned char tempByte;
	tempByte = ReadUART2();
	
#if defined API_MODE_2
	if (tempByte == 0x7D){
		WAIT_TIMEOUT
		tempByte = ReadUART2()^0x20;
	}
#endif

	return tempByte;
}

static void initPacket(unsigned char outPacket[], unsigned int *pIndex){
	// For packet structure, see XBee User's Manual 3.4.2 (API Types): TX Request (16-bit address)
	
	unsigned int packet_length;
	packet_length = ((DATA_ARRAY_LENGTH+ADDITIONAL_NUMS)*DATATYPE_BYTELENGTH+5);
	//	Packet length in bytes: 
	//	Add 5 bytes to data length for: 
	//	1. API identifier
	//	2. Frame ID
	//	3. Destination address high byte
	//	4. Destination address low byte
	//	5. Options byte	
	
	*pIndex=0;
	checksum=0;	
	outPacket[*pIndex]=0x7E;
	*pIndex+=1;
	outPacket[*pIndex]=packet_length/256; 
	//Packet length,high byte.  
	
	
#if defined API_MODE_2
	if (outPacket[*pIndex] == 0x7E ||
		outPacket[*pIndex] == 0x7D ||
		outPacket[*pIndex] == 0x11 ||
		outPacket[*pIndex] == 0x13){	
		outPacket[*pIndex+1] = outPacket[*pIndex]^0x20;
		outPacket[*pIndex] = 0x7D;
		*pIndex+=1;
	}
#endif
	
	*pIndex+=1;
	outPacket[*pIndex]=packet_length%256;    //Packet length, low byte
	
#if defined API_MODE_2
	if (outPacket[*pIndex] == 0x7E ||
		outPacket[*pIndex] == 0x7D ||
		outPacket[*pIndex] == 0x11 ||
		outPacket[*pIndex] == 0x13){	
		outPacket[*pIndex+1] = outPacket[*pIndex]^0x20;
		outPacket[*pIndex] = 0x7D;
		*pIndex+=1;
	}
#endif
	*pIndex+=1;
}

static void loadByte(unsigned char outPacket[],unsigned char byte, unsigned int *pIndex){
	outPacket[*pIndex]=byte;
	checksum+=outPacket[*pIndex];
	
#if defined API_MODE_2
	if (outPacket[*pIndex] == 0x7E ||
		outPacket[*pIndex] == 0x7D ||
		outPacket[*pIndex] == 0x11 ||
		outPacket[*pIndex] == 0x13){	
		outPacket[*pIndex+1] = outPacket[*pIndex]^0x20;
		outPacket[*pIndex] = 0x7D;
		*pIndex+=1;
	}
#endif
	*pIndex+=1;
}


static void sendPacket(unsigned char outPacket[],unsigned int Index){
	//index is the index of the char array, not the length of the array
	int i;
	for (i=0; i<Index; i++){ 
		while(BusyUART2()){}
		WriteUART2(outPacket[i]);
	}
}

#endif

